from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from tenacity import RetryError, wait_none

from alist_mikananirss import TaskMonitor
from alist_mikananirss.alist import Alist
from alist_mikananirss.alist.tasks import (
    AlistDownloadTask,
    AlistTaskList,
    AlistTaskStatus,
    AlistTaskType,
)


@pytest.fixture
def mock_alist():
    return AsyncMock(spec=Alist)


@pytest.fixture
def mock_task():
    return MagicMock(
        spec=AlistDownloadTask, tid="123", task_type=AlistTaskType.DOWNLOAD
    )


@pytest.fixture
def task_monitor(mock_alist, mock_task):
    return TaskMonitor(mock_alist, mock_task)


@pytest.mark.asyncio
async def test_refresh_success(task_monitor, mock_alist, mock_task):
    updated_task = MagicMock(
        spec=AlistDownloadTask, tid="123", status=AlistTaskStatus.Running, progress=0.5
    )
    mock_alist.get_task_list.return_value = AlistTaskList([updated_task])

    await task_monitor._refresh()

    mock_alist.get_task_list.assert_called_once_with(mock_task.task_type)
    assert task_monitor.task == updated_task


@pytest.mark.asyncio
async def test_refresh_task_not_found(task_monitor, mock_alist):
    mock_alist.get_task_list.return_value = AlistTaskList()

    with pytest.raises(RetryError):
        task_monitor._refresh.retry.wait = wait_none()
        await task_monitor._refresh()


@pytest.mark.asyncio
async def test_wait_finished_success(task_monitor):
    progresses = [0, 50, 100]
    mock_tasks = [
        MagicMock(
            spec=AlistDownloadTask,
            tid="123",
            status=AlistTaskStatus.Running,
            progress=p,
        )
        for p in progresses
    ]
    mock_tasks[-1].status = AlistTaskStatus.Succeeded
    success_task = mock_tasks[-1]

    async def mock_refresh():
        task_monitor.task = mock_tasks.pop(0)

    with (
        patch.object(task_monitor, "_refresh", side_effect=mock_refresh),
        patch("asyncio.sleep", new_callable=AsyncMock),
        patch("time.time", side_effect=[0, 1, 2, 3]),
    ):
        result = await task_monitor.wait_finished()

    assert result == success_task
    assert result.status == AlistTaskStatus.Succeeded


@pytest.mark.asyncio
async def test_wait_finished_failed(task_monitor):
    mock_task_failed = MagicMock(
        spec=AlistDownloadTask, tid="123", status=AlistTaskStatus.Failed, progress=0.5
    )

    async def mock_refresh():
        task_monitor.task = mock_task_failed

    with (
        patch.object(
            task_monitor,
            "_refresh",
            side_effect=mock_refresh,
        ),
        patch("asyncio.sleep", new_callable=AsyncMock),
    ):
        result = await task_monitor.wait_finished()

    assert result == mock_task_failed
    assert result.status == AlistTaskStatus.Failed

@pytest.mark.parametrize("task_type", [AlistTaskType.DOWNLOAD, AlistTaskType.TRANSFER])
def test_normal_status(task_type):
    assert set(TaskMonitor.NORMAL_STATUS) == {
        AlistTaskStatus.Pending,
        AlistTaskStatus.Running,
        AlistTaskStatus.StateBeforeRetry,
        AlistTaskStatus.StateWaitingRetry,
    }



import pytest

from alist_mikananirss.alist.api import AlistDownloaderType
from alist_mikananirss.common.config import ConfigManager


def test_load_config():
    config_path = "tests/common/test_config_valid.yaml"
    ConfigManager().load_config(config_path)
    cfg = ConfigManager().get_config()
    assert cfg.common_interval_time == 300
    assert cfg.common_proxies is None

    assert cfg.alist_base_url == "http://127.0.0.1:5244"
    assert cfg.alist_token == "your_token"
    assert cfg.alist_downloader == AlistDownloaderType.ARIA
    assert cfg.alist_download_path == "Onedrive/Anime"

    assert isinstance(cfg.mikan_subscribe_url, list)
    assert all(
        filter_name in cfg.mikan_regex_pattern for filter_name in cfg.mikan_filters
    )

    assert cfg.notification_enable is True
    assert cfg.notification_telegram_enable is True
    assert cfg.notification_pushplus_enable is False
    assert cfg.notification_interval_time == 300

    assert cfg.rename_enable is True
    assert cfg.rename_chatgpt_api_key == "sk-xxx"
    assert cfg.rename_chatgpt_base_url == "https://example.com/v1"
    assert cfg.rename_chatgpt_model == "gpt-4o-mini"
    assert cfg.rename_format == "{name} S{season:02d}E{episode:02d} {fansub}"
    assert cfg.rename_enable is True
    assert cfg.rename_remap_cfg_path == "remap.yaml"

    assert cfg.bot_assistant_enable is False

    assert cfg.dev_log_level == "INFO"


def test_load_config_with_error():
    config_path = "tests/common/test_config_invalid.yaml"
    with pytest.raises(KeyError) as excinfo:
        ConfigManager().load_config(config_path)
    assert "alist.token is not found in config file" in str(excinfo.value)



import os
import secrets
import string
from copy import deepcopy
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from tenacity import wait_none, RetryError

from alist_mikananirss import AnimeDownloadTaskInfo, DownloadManager, SubscribeDatabase
from alist_mikananirss.alist import (
    Alist,
    AlistClientError,
    AlistDownloadTask,
    AlistTaskList,
    AlistTaskStatus,
    AlistTransferTask,
)
from alist_mikananirss.websites.models import LanguageType, ResourceInfo


@pytest.fixture
def download_path():
    return "/base/path"


def get_incomplete_download_task(resource):
    test_instance = DownloadManager()
    alphabet = "_-" + string.digits + string.ascii_letters
    tid = "".join(secrets.choice(alphabet) for _ in range(21))
    download_path = test_instance._build_download_path(resource).replace(os.sep, "/")
    json_data = {
        "error": "",
        "id": tid,
        "name": f"download {resource.torrent_url} to ({download_path})",
        "progress": 50.0,
        "state": 1,
        "status": "offline downloading",
    }

    return AlistDownloadTask.from_json(json_data)


def get_complete_download_task(download_task):
    new_task = deepcopy(download_task)
    new_task.status = AlistTaskStatus.Succeeded
    new_task.progress = 100
    return new_task


def get_incomplete_transfer_task(resource):
    test_instance = DownloadManager()
    alphabet = "_-" + string.digits + string.ascii_letters
    tid = "".join(secrets.choice(alphabet) for _ in range(21))
    uuid = "".join(
        secrets.choice(string.digits + string.ascii_lowercase) for _ in range(36)
    )
    download_path = test_instance._build_download_path(resource).replace(os.sep, "/")

    cleaned_path = download_path.strip("/")
    parts = cleaned_path.split("/", 1)
    tgt_drive = "/" + parts[0]
    drive_subdir = ""
    if len(parts) > 1:
        drive_subdir = "/" + parts[1]

    json_data = {
        "error": "",
        "id": tid,
        "name": f"transfer [](/path/to/alist/data/temp/qBittorrent/{uuid}/subfolder/{resource.anime_name}/filename.mp4) to [{tgt_drive}]({drive_subdir})",
        "progress": 50.0,
        "state": 1,
        "status": "running",
    }
    return AlistTransferTask.from_json(json_data)


def get_complete_transfer_task(transfer_task):
    new_task = deepcopy(transfer_task)
    new_task.status = AlistTaskStatus.Succeeded
    new_task.progress = 100
    return new_task


@pytest.fixture
def test_resources():
    resources = [
        # basic info
        ResourceInfo(
            resource_title="title1",
            torrent_url="https://test1.torrent",
        ),
        # only anime name
        ResourceInfo(
            resource_title="title2",
            torrent_url="https://test2.torrent",
            anime_name="Test Anime",
        ),
        # full info
        ResourceInfo(
            resource_title="title3",
            torrent_url="https://test3.torrent",
            published_date="1",
            anime_name="Test Anime",
            season=1,
            episode=5,
            fansub="TestSub",
            quality="1080p",
            languages=[LanguageType.SIMPLIFIED_CHINESE],
        ),
    ]
    return resources


@pytest.fixture
def download_manager(download_path):
    mock_alist = AsyncMock(spec=Alist)
    mock_db = AsyncMock(spec=SubscribeDatabase)
    DownloadManager.initialize(
        mock_alist, download_path, use_renamer=True, need_notification=True, db=mock_db
    )
    DownloadManager().alist_client = mock_alist
    DownloadManager().db = mock_db
    return DownloadManager()


@pytest.mark.asyncio
async def test_download_success(download_manager, test_resources):
    incomplete_download_tasks = []
    for resource in test_resources:
        incomplete_download_tasks.append(get_incomplete_download_task(resource))

    side_effect = []
    tasks_by_path = {}
    for task in incomplete_download_tasks:
        path = task.download_path
        if path not in tasks_by_path:
            tasks_by_path[path] = []
        tasks_by_path[path].append(task)

    for tasks in tasks_by_path.values():
        side_effect.append(AlistTaskList(tasks))

    download_manager.alist_client.add_offline_download_task.side_effect = side_effect

    with patch("asyncio.create_task", new_callable=MagicMock):
        with patch.object(download_manager, "monitor", new_callable=MagicMock):
            await DownloadManager.add_download_tasks(test_resources)

            assert (
                download_manager.alist_client.add_offline_download_task.call_count
                == len(side_effect)
            )
            assert download_manager.db.insert_resource_info.call_count == len(
                test_resources
            )
            calls = download_manager.db.insert_resource_info.call_args_list
            for i in range(len(test_resources)):
                assert calls[i].args[0] == test_resources[i]


@pytest.mark.asyncio
async def test_download_failed(download_manager, test_resources):
    incomplete_download_tasks = []
    for resource in test_resources:
        incomplete_download_tasks.append(get_incomplete_download_task(resource))

    task_lists = []
    tasks_by_path = {}
    for task in incomplete_download_tasks:
        path = task.download_path
        if path not in tasks_by_path:
            tasks_by_path[path] = []
        tasks_by_path[path].append(task)

    for tasks in tasks_by_path.values():
        task_lists.append(AlistTaskList(tasks))
    side_effect = []
    for i in range(len(task_lists)):
        if i == 0:
            side_effect.append(task_lists[i])
        else:
            side_effect.append(TimeoutError)

    download_manager.alist_client.add_offline_download_task.side_effect = side_effect

    with patch("asyncio.create_task", new_callable=MagicMock):
        with patch.object(download_manager, "monitor", new_callable=MagicMock):
            await DownloadManager.add_download_tasks(test_resources)
            assert (
                download_manager.alist_client.add_offline_download_task.call_count
                == len(task_lists)
            )
            assert download_manager.db.insert_resource_info.call_count == 1
            calls = download_manager.db.insert_resource_info.call_args_list
            assert calls[0].args[0] == test_resources[0]


@pytest.mark.asyncio
async def test_monitor_success(download_manager, test_resources):
    incomplete_download_tasks = []
    complete_download_tasks = []
    incomplete_transfer_tasks = []
    complete_transfer_tasks = []

    for resource in test_resources:
        incomplete_download_tasks.append(get_incomplete_download_task(resource))
        complete_download_tasks.append(
            get_complete_download_task(incomplete_download_tasks[-1])
        )
        incomplete_transfer_tasks.append(get_incomplete_transfer_task(resource))
        complete_transfer_tasks.append(
            get_complete_transfer_task(incomplete_transfer_tasks[-1])
        )

    task_info_list = []
    for i in range(len(test_resources)):
        task_info_list.append(
            AnimeDownloadTaskInfo(
                resource=test_resources[i],
                download_task=incomplete_download_tasks[i],
            )
        )

    with (
        patch(
            "alist_mikananirss.core.download_manager.TaskMonitor"
        ) as mock_task_monitor,
        patch.object(download_manager, "_post_process") as mock_post_process,
        patch.object(
            download_manager, "_find_transfer_task"
        ) as mock_find_transfer_task,
    ):
        for i in range(len(test_resources)):
            dl_monitor = AsyncMock()
            dl_monitor.wait_finished.side_effect = [complete_download_tasks[i]]
            mock_find_transfer_task.return_value = incomplete_transfer_tasks[i]
            tf_monitor = AsyncMock()
            tf_monitor.wait_finished.side_effect = [complete_transfer_tasks[i]]
            mock_task_monitor.side_effect = [dl_monitor, tf_monitor]
            await download_manager.monitor(task_info_list[i])

        assert mock_post_process.call_count == len(test_resources)


@pytest.mark.asyncio
async def test_monitor_failed(download_manager, test_resources):
    incomplete_download_task = get_incomplete_download_task(test_resources[0])

    task_info = AnimeDownloadTaskInfo(
        resource=test_resources[0],
        download_task=incomplete_download_task,
    )

    with (
        patch(
            "alist_mikananirss.core.download_manager.TaskMonitor"
        ) as mock_task_monitor,
        patch.object(
            download_manager.db, "delete_by_resource_title"
        ) as mock_delete_by_resource_title,
    ):
        dl_monitor = AsyncMock()
        dl_monitor.wait_finished.side_effect = [TimeoutError]
        mock_task_monitor.return_value = dl_monitor
        await download_manager.monitor(task_info)

        assert mock_delete_by_resource_title.call_count == 1
        assert (
            mock_delete_by_resource_title.call_args[0][0]
            == test_resources[0].resource_title
        )


@pytest.mark.asyncio
async def test_find_transfer_task_success(download_manager, test_resources):
    download_tasks = []
    transfer_tasks = []

    for resource in test_resources:
        download_tasks.append(
            get_complete_download_task(get_incomplete_download_task(resource))
        )
        transfer_tasks.append(get_incomplete_transfer_task(resource))

    mock_task_list = AlistTaskList(transfer_tasks)
    download_manager.alist_client.get_task_list.return_value = mock_task_list

    for i, dl_task in enumerate(download_tasks):
        find_res = await download_manager._find_transfer_task(dl_task)
        assert find_res == transfer_tasks[i]
        assert transfer_tasks[i].uuid in download_manager.uuid_set


@pytest.mark.asyncio
async def test_find_transfer_task_not_found(download_manager, test_resources):
    download_task = get_complete_download_task(
        get_incomplete_download_task(test_resources[0])
    )

    download_manager.alist_client.get_task_list.return_value = AlistTaskList([])
    download_manager._find_transfer_task.retry.wait = wait_none()
    result = await download_manager._find_transfer_task(download_task)

    assert result is None


@pytest.mark.asyncio
async def test_find_transfer_task_api_error(download_manager, test_resources):
    download_task = get_complete_download_task(
        get_incomplete_download_task(test_resources[0])
    )

    download_manager.alist_client.get_task_list.side_effect = AlistClientError(
        "Network error"
    )
    download_manager._find_transfer_task.retry.wait = wait_none()
    result = await download_manager._find_transfer_task(download_task)

    assert result is None



from __future__ import annotations

import re
from abc import ABC
from dataclasses import dataclass, field
from enum import Enum
from typing import Optional


# https://github.com/alist-org/alist/blob/86b35ae5cfec400871072356fec4dea88303195d/pkg/task/task.go#L27
class AlistTaskStatus(Enum):
    Pending = 0
    Running = 1
    Succeeded = 2
    Canceling = 3
    Canceled = 4
    Errored = 5
    Failing = 6
    Failed = 7
    StateWaitingRetry = 8
    StateBeforeRetry = 9
    UNKNOWN = 10


class AlistDownloaderType(Enum):
    ARIA = "aria2"
    QBIT = "qBittorrent"


class AlistDeletePolicy(Enum):
    DeleteOnUploadSucceed = "delete_on_upload_succeed"
    DeleteOnUploadFailed = "delete_on_upload_failed"
    DeleteNever = "delete_never"
    DeleteAlways = "delete_always"


class AlistTaskType(Enum):
    DOWNLOAD = "offline_download"
    TRANSFER = "offline_download_transfer"
    UNKNOWN = "unknown"


# for task related api call
class AlistTaskState(Enum):
    DONE = "done"
    UNDONE = "undone"


DOWNLOAD_DES_PATTERN = re.compile(r"download\s+(.+?)\s+to \((.+?)\)")
TRANSFER_DES_PATTERN = re.compile(r"transfer \[.*\]\((.+)\) to \[(.+)\]\((.+)\)")


class AlistTaskError(Exception):
    """Base exception for AlistTask related errors."""


class InvalidTaskDescription(AlistTaskError):
    """Raised when task description is invalid."""


@dataclass
class AlistTask(ABC):
    tid: str
    description: str
    status: AlistTaskStatus
    status_msg: str
    progress: float
    error_msg: str = ""
    task_type: AlistTaskType = field(init=False)

    @classmethod
    def from_json(cls, json_data: dict) -> "AlistTask":
        """Creates an AlistTask instance from a JSON dictionary."""
        tid = json_data["id"]
        description = json_data["name"]
        status = AlistTaskStatus(json_data["state"])
        status_msg = json_data["status"]
        progress = json_data["progress"]
        error_msg = json_data.get("error")

        return cls(
            tid=tid,
            description=description,
            status=status,
            status_msg=status_msg,
            progress=progress,
            error_msg=error_msg,
        )


@dataclass
class AlistTransferTask(AlistTask):
    uuid: str = field(init=False)
    target_path: str = field(init=False)
    task_type: AlistTaskType = field(default=AlistTaskType.TRANSFER, init=False)

    def __post_init__(self):
        self.task_type = AlistTaskType.TRANSFER
        match = re.match(TRANSFER_DES_PATTERN, self.description)
        if match:
            temp_filepath = match.group(1)
            target_drive = match.group(2)
            drive_subdir = match.group(3)
            target_dirpath = f"{target_drive}{drive_subdir}"
            elements = temp_filepath.split("/")
            uuid = elements[elements.index("temp") + 2]
            sub_path = temp_filepath[temp_filepath.rfind(uuid) + len(uuid) + 1 :]
            target_file_path = f"{target_dirpath}/{sub_path}"
        else:
            raise InvalidTaskDescription(
                f"Failed to get uuid and target filepath from task description: {self.description}"
            )
        self.uuid = uuid
        self.target_path = target_file_path


@dataclass
class AlistDownloadTask(AlistTask):
    url: str = field(init=False)
    download_path: str = field(init=False)
    task_type: AlistTaskType = field(default=AlistTaskType.DOWNLOAD, init=False)

    def __post_init__(self):
        match = re.match(DOWNLOAD_DES_PATTERN, self.description)
        if match:
            url = match.group(1)
            download_path = match.group(2)
        else:
            raise InvalidTaskDescription(
                f"Failed to get url and download path from task description: {self.description}"
            )
        # If seeding, the task status will still be Running
        # We need to change it to Succeeded manually to ensure the task is marked as completed
        if (
            self.status == AlistTaskStatus.Running
            and "offline download completed" in self.status_msg
        ):
            self.status = AlistTaskStatus.Succeeded

        self.url = url
        self.download_path = download_path


class AlistTaskList:

    def __init__(self, tasks: list[AlistTask] = None):
        self.tasks = tasks or []
        self.id_map = {}
        for task in self.tasks:
            self.id_map[task.tid] = task

    def add_task(self, task: AlistTask):
        self.tasks.append(task)
        self.id_map[task.tid] = task

    def __iter__(self):
        return iter(self.tasks)

    def __add__(self, other: AlistTaskList):
        self.tasks.extend(other.tasks)
        self.id_map.update(other.id_map)
        return self

    def __getitem__(self, idx: int) -> AlistTask:
        return self.tasks[idx]

    def __len__(self) -> int:
        return len(self.tasks)

    def __contains__(self, tid: str) -> bool:
        return tid in self.id_map

    def __repr__(self):
        return f"{self.tasks}"

    def get_by_id(self, tid: str) -> Optional[AlistTask]:
        """Get task by id, return None if not found."""
        return self.id_map.get(tid, None)
